@@ -30,7 +30,7 @@ import os
 from glob import glob
 
 
-__version__ = '0.0.13'
+__version__ = '0.3.1'
 
 
 def create_app(test_config=None):
@@ -64,6 +64,7 @@ def create_app(test_config=None):
 
     app.config["IMAGES"] = os.path.join(app.instance_path, 'images')
     app.config["USER_IMAGES"] = os.path.join(app.config["IMAGES"], 'user')
+    app.config["ENTRY_FILES"] = os.path.join(app.instance_path, 'entry_files')
     app.config["ALLOWED_IMAGE_EXTENSIONS"] = ["JPEG", "JPG", "PNG", "GIF"]
     app.config["SESSIONS"] = os.path.join(app.instance_path, 'sessions')
     app.config["PAGES"] = os.path.join(app.instance_path, 'pages')
@@ -82,6 +83,7 @@ def create_app(test_config=None):
 
     # ensure the instance folder exists
     os.makedirs(app.config['USER_IMAGES'], exist_ok=True)
+    os.makedirs(app.config['ENTRY_FILES'], exist_ok=True)
     os.makedirs(app.config['SESSIONS'], exist_ok=True)
     os.makedirs(app.config['PAGES'], exist_ok=True)
 
@@ -104,6 +106,15 @@ def create_app(test_config=None):
                   'params': params
                   }
 
+        if 'login_user' in params and app.is_devel:
+            user_id = params['login_user']
+            user = app.dm.get_user_by(id=user_id)
+            if user is None:
+                return send_error("Invalid user id: '%s'" % user_id)
+            elif user != app.user:
+                flask_login.logout_user()
+                flask_login.login_user(user)
+
         if app.user.is_authenticated:
             if content_id == 'user_login':  # Redirects to Dashboard by default
                 kwargs['content_id'] = 'dashboard'
@@ -118,6 +129,7 @@ def create_app(test_config=None):
         kwargs['version'] = __version__
         kwargs['emhub_title'] = app.config.get('EMHUB_TITLE', '')
         kwargs['possible_owners'] = app.dc.get_pi_labs()
+        kwargs['possible_operators'] = app.dc.get_possible_operators()
         kwargs.update(app.dc.get_resources_list())
 
         return flask.render_template('main.html', **kwargs)
@@ -247,7 +259,16 @@ def create_app(test_config=None):
             try:
                 kwargs = app.dc.get(**content_kwargs)
             except Exception as e:
-                return "<h1><span style='color: red'>Error</span> </br>%s</h1>" % e
+                import traceback
+                tb = traceback.format_exc()
+                error = {
+                    'message': str(e),
+                    'body': tb
+                }
+                return flask.render_template('error_dialog.html', error=error)
+                # result = "<div><h1><span style='color: red'>Error</span> </br>%s</h1>" % e
+                # result += "<pre>%s</pre></div>" % tb
+                # return result
         else:
             kwargs = {'next_content': content_id}
             content_id = 'user_login'
@@ -258,7 +279,10 @@ def create_app(test_config=None):
             kwargs['is_devel'] = app.is_devel
             return flask.render_template(content_template, **kwargs)
 
-        return "<h1>Template '%s' not found</h1>" % content_template
+        error = {
+            "message": "Template '%s' not found." % content_template
+        }
+        return flask.render_template('error_dialog.html', error=error)
 
 
     @app.template_filter('basename')
@@ -267,7 +291,8 @@ def create_app(test_config=None):
 
     @app.template_filter('id_from_label')
     def id_from_label(label):
-        return label.replace(' ', '_')
+        return label.translate(label.maketrans("", "", "!#$%^&*()/\\ "))
+        #return label.replace(' ', '_')
 
     @app.template_filter('range_params')
     def range_params(date_range):
@@ -275,6 +300,11 @@ def create_app(test_config=None):
         e = date_range[1].strftime("%Y/%m/%d")
         return "&start=%s&end=%s" % (s, e)
 
+    def url_for_content(contentId, **kwargs):
+        return flask.url_for('main', content_id=contentId, **kwargs)
+
+    app.jinja_env.globals.update(url_for_content=url_for_content)
+    app.jinja_env.add_extension('jinja2.ext.do')
     app.jinja_env.filters['reverse'] = basename
     app.jinja_env.filters['pretty_datetime'] = pretty_datetime
     app.jinja_env.filters['pretty_date'] = pretty_date
