@@ -27,13 +27,17 @@
 # **************************************************************************
 
 import os
+import time
+import json
+from glob import glob
 
 import flask
 from flask import request
 from flask import current_app as app
 import flask_login
 
-from emhub.utils import datetime_from_isoformat, send_json_data, send_error
+from emhub.utils import (datetime_from_isoformat, datetime_to_isoformat,
+                         send_json_data, send_error)
 from emhub.data import DataContent
 
 
@@ -41,6 +45,7 @@ api_bp = flask.Blueprint('api', __name__)
 
 
 # ---------------------------- AUTH  ------------------------------------------
+
 @api_bp.route('/login', methods=['POST'])
 def login():
     username = request.json['username']
@@ -48,7 +53,7 @@ def login():
 
     user = app.dm.get_user_by(username=username)
     if user is None or not user.check_password(password):
-        send_error('Invalid username or password')
+        return send_error('Invalid username or password')
 
     flask_login.login_user(user)
 
@@ -64,6 +69,7 @@ def logout():
 
 
 # ---------------------------- USERS ------------------------------------------
+
 @api_bp.route('/create_user', methods=['POST'])
 @flask_login.login_required
 def create_user():
@@ -77,22 +83,20 @@ def update_user():
         f = request.form
 
         attrs = {'id': f['user-id'],
-                 #'username': f['user-username'],
                  'name': f['user-name'],
-                 'phone': f['user-phone'],
-                 'roles': [v.replace('role-', '')
-                           for v in f if v.startswith('role-')]
+                 'phone': f['user-phone']
                  }
 
+        roles = [v.replace('role-', '') for v in f if v.startswith('role-')]
+        if roles:
+            attrs['roles'] = roles
+
         if 'user-pi-select' in f:
             attrs['pi_id'] = int(f['user-pi-select'])
             # TODO: Validate if a user is not longer PI
             # check that there are not other users referencing this one as pi
             # still this will not be a very common case
 
-        for v in f:
-            print(v, '=', f[v])
-
         password = f['user-password'].strip()
         if password:
             attrs['password'] = password
@@ -127,6 +131,7 @@ def get_users():
 
 
 # ---------------------------- APPLICATIONS -----------------------------------
+
 @api_bp.route('/create_template', methods=['POST'])
 @flask_login.login_required
 def create_template():
@@ -154,7 +159,7 @@ def delete_template():
 @api_bp.route('/create_application', methods=['POST'])
 @flask_login.login_required
 def create_application():
-    return send_error('create_application NOT IMPLEMENTED')
+    return handle_template(app.dm.create_application)
 
 
 @api_bp.route('/get_applications', methods=['POST'])
@@ -169,47 +174,97 @@ def update_application():
     return handle_application(app.dm.update_application)
 
 
-@api_bp.route('/add_application_users', methods=['POST'])
+@api_bp.route('/import_application', methods=['POST'])
 @flask_login.login_required
-def add_application_users():
+def import_application():
     try:
         if not request.json:
             raise Exception("Expecting JSON request.")
 
-        app_id = request.json['application_id']
-        users = request.json['users']
-        if isinstance(users, str):
-            users = map(int, users.split())
+        orderCode = request.json['code'].upper()
 
-        application = app.dm.get_application_by(id=app_id)
+        dm = app.dm
 
-        errorMsg = ''
+        application = dm.get_application_by(code=orderCode)
 
-        if application is None:
-            errorMsg += "Invalid application id: %s" % app_id
+        if application is not None:
+            raise Exception('Application %s already exist' % orderCode)
 
-        for u in users:
-            user = app.dm.get_user_by(id=u)
-            if user is None:
-                errorMsg += "\nInvalid user id: %s" % u
-            else:
-                application.users.append(user)
+        orderJson = app.sll_pm.fetchOrderDetailsJson(orderCode)
 
-        if errorMsg:
-            raise Exception(errorMsg)
+        if orderJson is None:
+            raise Exception('Invalid application ID %s' % orderCode)
 
-        app.dm.commit()
-        return send_json_data({'OK': True})
+        piEmail = orderJson['owner']['email'].lower()
+        # orderId = orderJson['identifier']
+
+        pi = dm.get_user_by(email=piEmail)
+
+        if pi is None:
+            raise Exception("Order owner email (%s) not found as PI" % piEmail)
+
+        if orderJson['status'] not in ['accepted', 'processing']:
+            raise Exception("Only applications with status 'accepted' or "
+                            "'processing' can be imported. ")
+
+        fields = orderJson['fields']
+        description = fields.get('project_des', None)
+        invoiceRef = fields.get('project_invoice_addess', None)
+
+        created = datetime_from_isoformat(orderJson['created'])
+        pi_list = fields.get('pi_list', [])
+
+        form = orderJson['form']
+        iuid = form['iuid']
+
+        # Check if the given form (here templates) already exist
+        # or we need to create a new one
+        orderTemplate = None
+        templates = dm.get_templates()
+        for t in templates:
+            if t.extra.get('portal_iuid', None) == iuid:
+                orderTemplate = t
+                break
+
+        if orderTemplate is None:
+            orderTemplate = dm.create_template(
+                title=form['title'],
+                status='active',
+                extra={'portal_iuid': iuid}
+            )
+            dm.commit()
+
+        application = dm.create_application(
+            code=orderCode,
+            title=orderJson['title'],
+            created=created,  # datetime_from_isoformat(o['created']),
+            status='active',
+            description=description,
+            creator_id=pi.id,
+            template_id=orderTemplate.id,
+            invoice_reference=invoiceRef or 'MISSING_INVOICE_REF',
+        )
+
+        for piTuple in pi_list:
+            piEmail = piTuple[1].lower()
+            pi = dm.get_user_by(email=piEmail)
+            if pi is not None:
+                application.users.append(pi)
+
+        dm.commit()
+
+        return send_json_data({'application': application.json()})
 
     except Exception as e:
         print(e)
         import traceback
         traceback.print_exc()
+
         return send_error('ERROR from Server: %s' % e)
 
 
-
 # ---------------------------- RESOURCES ---------------------------------------
+
 @api_bp.route('/get_resources', methods=['POST'])
 @flask_login.login_required
 def get_resources():
@@ -221,14 +276,15 @@ def get_resources():
 def update_resource():
     return handle_resource(app.dm.update_resource)
 
+
 # ---------------------------- BOOKINGS ---------------------------------------
 
 @api_bp.route('/create_booking', methods=['POST'])
 @flask_login.login_required
 def create_booking():
     def create(**attrs):
-        check_min = request.json.get('check_min_booking')
-        check_max = request.json.get('check_max_booking')
+        check_min = request.json.get('check_min_booking', True)
+        check_max = request.json.get('check_max_booking', True)
         return app.dm.create_booking(check_min_booking=check_min,
                                      check_max_booking=check_max,
                                      **attrs)
@@ -280,6 +336,40 @@ def get_sessions():
     return filter_request(app.dm.get_sessions)
 
 
+@api_bp.route('/poll_sessions', methods=['POST'])
+@flask_login.login_required
+def poll_sessions():
+    session_folders = app.dm.get_session_folders()
+
+    def _user(u):
+        if not u:
+            return {}
+        else:
+            return {'name': u.name,
+                    'email': u.email
+                    }
+
+    while True:
+        sessions = app.dm.get_sessions(condition='status=="pending"')
+        if sessions:
+            for s in sessions:
+                b = s.booking
+                e = app.dc.booking_to_event(b)
+                data = [{
+                    'id': s.id,
+                    'name': s.name,
+                    'booking_id': s.booking_id,
+                    'start': datetime_to_isoformat(s.start),
+                    'user': _user(b.owner),
+                    'pi': _user(b.owner.get_pi()),
+                    'operator': _user(b.operator),
+                    'folder': session_folders[s.name[:3]],
+                    'title': e['title']
+                 }]
+                return send_json_data(data)
+        time.sleep(3)
+
+
 @api_bp.route('/create_session', methods=['POST'])
 @flask_login.login_required
 def create_session():
@@ -304,18 +394,46 @@ def load_session():
     return handle_session(app.dm.load_session)
 
 
+@api_bp.route('/clear_session_data', methods=['POST'])
+@flask_login.login_required
+def clear_session_data():
+    return handle_session(app.dm.clear_session_data)
+
+
 @api_bp.route('/create_session_set', methods=['POST'])
 @flask_login.login_required
 def create_session_set():
     """ Create a set file without actual session. """
     def handle(session, set_id, **attrs):
         session.data.create_set(set_id, attrs)
-        session.data.close()
         return {'session_set': {}}
 
     return handle_session_data(handle, mode="a")
 
 
+@api_bp.route('/update_session_set', methods=['POST'])
+@flask_login.login_required
+def update_session_set():
+    """ Create a set file without actual session. """
+    def handle(session, set_id, **attrs):
+        session.data.update_set(set_id, attrs)
+        return {'session_set': {}}
+
+    return handle_session_data(handle, mode="a")
+
+
+@api_bp.route('/get_session_sets', methods=['POST'])
+@flask_login.login_required
+def get_session_sets():
+    """ Return all sets' name of this session. """
+    def handle(session, set_id, **attrs):
+        # set_id is not used here, but passed by default to the handle
+        sets = session.data.get_sets(attrList=attrs)
+        return {'session_sets': sets}
+
+    return handle_session_data(handle, mode="r")
+
+
 @api_bp.route('/add_session_item', methods=['POST'])
 @flask_login.login_required
 def add_session_item():
@@ -323,7 +441,6 @@ def add_session_item():
     def handle(session, set_id, **attrs):
         itemId = attrs.pop("item_id")
         session.data.add_set_item(set_id, itemId, attrs)
-        session.data.close()
         return {'item': {}}
 
     return handle_session_data(handle, mode="a")
@@ -335,8 +452,7 @@ def update_session_item():
     """ Update existing item. """
     def handle(session, set_id, **attrs):
         itemId = attrs.pop("item_id")
-        session.data.update_item(set_id, itemId, attrs)
-        session.data.close()
+        session.data.update_set_item(set_id, int(itemId), attrs)
         return {'item': {}}
 
     return handle_session_data(handle, mode="a")
@@ -347,9 +463,7 @@ def update_session_item():
 def get_session_data():
     """ Return some information related to session (e.g CTF values, etc). """
     def handle(session, set_id, **attrs):
-        result = DataContent(app).get_session_data(session)
-        session.data.close()
-        return result
+        return DataContent(app).get_session_data(session)
 
     return handle_session_data(handle, mode="r")
 
@@ -406,6 +520,134 @@ def delete_transaction():
     return handle_transaction(app.dm.delete_transaction)
 
 
+# ------------------------------ FORMS ---------------------------------
+
+@api_bp.route('/get_forms', methods=['GET', 'POST'])
+@flask_login.login_required
+def get_forms():
+    return send_json_data([f.json() for f in app.dm.get_forms()])
+
+
+@api_bp.route('/create_form', methods=['POST'])
+@flask_login.login_required
+def create_form():
+    return handle_form(app.dm.create_form)
+
+
+@api_bp.route('/update_form', methods=['POST'])
+@flask_login.login_required
+def update_form():
+    return handle_form(app.dm.update_form)
+
+
+@api_bp.route('/delete_form', methods=['POST'])
+@flask_login.login_required
+def delete_form():
+    return handle_form(app.dm.delete_form)
+
+
+# ------------------------------ PROJECTS ---------------------------------
+
+@api_bp.route('/get_projects', methods=['GET', 'POST'])
+@flask_login.login_required
+def get_projects():
+    return send_json_data([p.json() for p in app.dm.get_projects()])
+
+
+@api_bp.route('/create_project', methods=['POST'])
+@flask_login.login_required
+def create_project():
+    return handle_project(app.dm.create_project)
+
+
+@api_bp.route('/update_project', methods=['POST'])
+@flask_login.login_required
+def update_project():
+    return handle_project(app.dm.update_project)
+
+
+@api_bp.route('/delete_project', methods=['POST'])
+@flask_login.login_required
+def delete_project():
+    return handle_project(app.dm.delete_project)
+
+
+# ------------------------------ ENTRIES ---------------------------------
+
+@api_bp.route('/get_entries', methods=['GET', 'POST'])
+@flask_login.login_required
+def get_entries():
+    return send_json_data([p.json() for p in app.dm.get_entries()])
+
+
+@api_bp.route('/create_entry', methods=['POST'])
+@flask_login.login_required
+def create_entry():
+    def handle(**attrs):
+        fix_dates(attrs, 'date')
+        entry = app.dm.create_entry(**attrs)
+        save_entry_files(entry, entry.extra['data'])
+        entry = app.dm.update_entry(id=entry.id, extra=entry.extra)
+        return entry.json()
+
+    return _handle_item(handle, 'entry')
+
+
+@api_bp.route('/update_entry', methods=['POST'])
+@flask_login.login_required
+def update_entry():
+    def handle(**attrs):
+        fix_dates(attrs, 'date')
+        entry = app.dm.get_entry_by(id=attrs['id'])
+        old_files = set(app.dm.get_entry_files(entry))
+        save_entry_files(entry, attrs['extra']['data'])
+        entry = app.dm.update_entry(**attrs)
+        new_files = set(app.dm.get_entry_files(entry))
+        clean_files(old_files - new_files)
+        return entry.json()
+
+    return _handle_item(handle, 'entry')
+
+
+@api_bp.route('/delete_entry', methods=['POST'])
+@flask_login.login_required
+def delete_entry():
+    def handle(**attrs):
+        entry = app.dm.get_entry_by(id=attrs['id'])
+        old_files = set(app.dm.get_entry_files(entry))
+        entry = app.dm.delete_entry(**attrs)
+        clean_files(old_files)
+        return entry.json()
+
+    return _handle_item(handle, 'entry')
+
+
+# ------------------------------ ENTRIES ---------------------------------
+
+@api_bp.route('/get_entries', methods=['GET', 'POST'])
+@flask_login.login_required
+def get_pucks():
+    return send_json_data([p.json() for p in app.dm.get_pucks()])
+
+
+@api_bp.route('/create_puck', methods=['POST'])
+@flask_login.login_required
+def create_puck():
+    return handle_puck(app.dm.create_puck,)
+
+
+@api_bp.route('/update_puck', methods=['POST'])
+@flask_login.login_required
+def update_puck():
+    return handle_puck(app.dm.update_puck)
+
+
+@api_bp.route('/delete_puck', methods=['POST'])
+@flask_login.login_required
+def delete_puck():
+    return handle_puck(app.dm.delete_puck)
+
+
 # -------------------- UTILS functions ----------------------------------------
 
 def filter_request(func):
@@ -426,11 +668,23 @@ def filter_request(func):
     return send_json_data(items)
 
 
+def fix_dates(attrs, *date_keys):
+    """ Convert the values from UTC string to datetime
+    for some keys that might be present in the attrs dict. """
+    for date_key in date_keys:
+        if date_key in attrs:
+            attrs[date_key] = datetime_from_isoformat(attrs[date_key])
+
+
 def _handle_item(handle_func, result_key):
     try:
-        if not request.json:
-            raise Exception("Expecting JSON request.")
-        result = handle_func(**request.json['attrs'])
+        if request.json:
+            attrs = request.json['attrs']
+        elif request.form:
+            attrs = json.loads(request.form['attrs'])
+        else:
+            raise Exception("Expecting JSON or Form request.")
+        result = handle_func(**attrs)
         return send_json_data({result_key: result})
     except Exception as e:
         print(e)
@@ -441,15 +695,12 @@ def _handle_item(handle_func, result_key):
 
 def handle_booking(result_key, booking_func, booking_transform=None):
     def handle(**attrs):
-        def _fix_date(date_key):
-            if date_key in attrs:
-                attrs[date_key] = datetime_from_isoformat(attrs[date_key])
-
-        _fix_date('start')
-        _fix_date('end')
+        dates = ['start', 'end']
 
         if attrs.get('repeat_value', 'no') != 'no':
-            _fix_date('repeat_stop')
+            dates.append('repeat_stop')
+
+        fix_dates(attrs, *dates)
 
         bt = booking_transform or app.dc.booking_to_event
         return [bt(b) for b in booking_func(**attrs)]
@@ -480,12 +731,7 @@ def handle_resource(resource_func):
 
 def handle_session(session_func):
     def handle(**attrs):
-        def _fix_date(date_key):
-            if date_key in attrs:
-                attrs[date_key] = datetime_from_isoformat(attrs[date_key])
-
-        _fix_date('start')
-        _fix_date('end')
+        fix_dates(attrs, 'start', 'end')
         return session_func(**attrs).json()
 
     return _handle_item(handle, 'session')
@@ -494,22 +740,27 @@ def handle_session(session_func):
 def handle_session_data(handle, mode="r"):
     attrs = request.json['attrs']
     session_id = attrs.pop("session_id")
-    set_id = attrs.pop("set_id", 1)
+    set_id = attrs.pop("set_id", None)
+    tries = 0
 
-    session = app.dm.load_session(sessionId=session_id, mode=mode)
-    result = handle(session, set_id, **attrs)
+    while tries < 3:
+        try:
+            session = app.dm.load_session(sessionId=session_id, mode=mode)
+            result = handle(session, set_id, **attrs)
+            session.data.close()
+            break
+        except OSError:
+            print("Error with session data, sleeping 3 secs")
+            time.sleep(3)
+            result = {}
+            tries += 1
 
     return send_json_data(result)
 
 
 def handle_invoice_period(invoice_period_func):
     def handle(**attrs):
-        def _fix_date(date_key):
-            if date_key in attrs:
-                attrs[date_key] = datetime_from_isoformat(attrs[date_key])
-
-        _fix_date('start')
-        _fix_date('end')
+        fix_dates(attrs, 'start', 'end')
         return invoice_period_func(**attrs).json()
 
     return _handle_item(handle, 'invoice_period')
@@ -517,19 +768,53 @@ def handle_invoice_period(invoice_period_func):
 
 def handle_transaction(transaction_func):
     def handle(**attrs):
-
-        print("handle_transaction: ", attrs)
-
-        def _fix_date(date_key):
-            if date_key in attrs:
-                attrs[date_key] = datetime_from_isoformat(attrs[date_key])
-
-        _fix_date('date')
+        fix_dates(attrs, 'date')
         return transaction_func(**attrs).json()
 
     return _handle_item(handle, 'transaction')
 
 
+def handle_form(form_func):
+    def handle(**attrs):
+        return form_func(**attrs).json()
+
+    return _handle_item(handle, 'form')
+
+
+def handle_project(project_func):
+    def handle(**attrs):
+        fix_dates(attrs, 'date')
+        return project_func(**attrs).json()
+
+    return _handle_item(handle, 'project')
+
+
+def save_entry_files(entry, data):
+    """ Function used when creating or updating an Entry.
+    It will handle the upload of new files.
+    """
+    for f in request.files:
+        file = request.files[f]
+
+        fn = file.filename
+        if fn:
+            data[f] = fn
+            file.save(app.dm.get_entry_path(entry, fn))
+
+
+def clean_files(paths):
+    for p in paths:
+        if os.path.exists(p):
+            os.remove(p)
+
+
+def handle_puck(entry_func):
+    def handle(**attrs):
+        return entry_func(**attrs).json()
+
+    return _handle_item(handle, 'entry')
+
+
 def create_item(name):
     def handle(**attrs):
         create_func = getattr(app.dm, 'create_%s' % name)
