@@ -82,7 +82,7 @@ def create_data_models(dm):
             return self.extra.get(key, default)
 
         def __setExtra(self, key, value):
-            extra = dict(self.extra)
+            extra = dict(self.extra or {})
             extra[key] = value
             self.extra = extra
 
@@ -121,7 +121,7 @@ def create_data_models(dm):
 
         @property
         def max_booking(self):
-            """ Minimum amount of hours that should be used for
+            """ Maximum amount of hours that can be used in a booking for
             booking this resource.
             """
             return self.__getExtra('max_booking', 0)
@@ -170,10 +170,11 @@ def create_data_models(dm):
                             Column('user_id', Integer,
                                    ForeignKey('users.id')))
 
+
     class User(UserMixin, Base):
         """Model for user accounts."""
         __tablename__ = 'users'
-        ROLES = ['user', 'admin', 'manager', 'head', 'pi']
+        ROLES = ['user', 'admin', 'manager', 'head', 'pi', 'independent']
 
         id = Column(Integer,
                     primary_key=True)
@@ -238,6 +239,8 @@ def create_data_models(dm):
 
         transactions = relationship('Transaction', back_populates="user")
 
+        #projects = relationship('Project', back_populates="user")
+
         # General JSON dict to store extra attributes
         extra = Column(JSON, default={})
 
@@ -301,6 +304,10 @@ def create_data_models(dm):
         def is_pi(self):
             return 'pi' in self.roles
 
+        @property
+        def is_independent(self):
+            return 'independent' in self.roles
+
         @property
         def is_application_manager(self):
             return len(self.created_applications) > 0
@@ -321,7 +328,7 @@ def create_data_models(dm):
 
         def same_pi(self, other):
             """ Return if the same pi. """
-            return self.get_pi() == other.get_pi()
+            return other is not None and self.get_pi() == other.get_pi()
 
         def get_applications(self, status='active'):
             """ Return the applications of this user.
@@ -361,6 +368,7 @@ def create_data_models(dm):
             """ Return True if the user can book in the given SLOT. """
             return booking_slot.allows_user_in_slot(self)
 
+
     class Template(Base):
         """ Classes used as template to create Applications.
         Template instances that are 'active' will allow to
@@ -390,9 +398,29 @@ def create_data_models(dm):
         # General JSON dict to store extra attributes
         extra = Column(JSON, default={})
 
+        def __getExtra(self, key, default):
+            return self.extra.get(key, default)
+
+        def __setExtra(self, key, value):
+            extra = dict(self.extra)
+            extra[key] = value
+            self.extra = extra
+
+        @property
+        def code_prefix(self):
+            """ Prefix for numbering Application Codes
+            (e.g CEM, EXT, etc)
+            """
+            return  self.__getExtra('code_prefix', [])
+
+        @code_prefix.setter
+        def codes(self, value):
+            self.__setExtra('code_prefix', value)
+
         def json(self):
             return dm.json_from_object(self)
 
+
     class Application(Base):
         """
         Application that applies for access to the facility.
@@ -466,6 +494,24 @@ def create_data_models(dm):
         def __repr__(self):
             return '<Application code=%s, alias=%s>' % (self.code, self.alias)
 
+        def __getExtra(self, key, default):
+            return self.extra.get(key, default)
+
+        def __setExtra(self, key, value):
+            extra = dict(self.extra)
+            extra[key] = value
+            self.extra = extra
+
+        @property
+        def confidential(self):
+            """ Return extra costs associated with this Booking
+            """
+            return  self.__getExtra('confidential', False)
+
+        @confidential.setter
+        def costs(self, value):
+            self.__setExtra('confidential', value)
+
         def json(self):
             json = dm.json_from_object(self)
             json['pi_list'] = [pi.id for pi in self.pi_list]
@@ -500,37 +546,41 @@ def create_data_models(dm):
             # Since we are importing data now from the Portal, some application
             # have the creator of the application as one of the users, so we want
             # to avoid duplicated entries
-            pi_list = [self.creator]
+            pi_list = []
+
+            # Applications can also be created by facility staff, so only if the
+            # creator is a pi we reported in the "pi_list"
+            if self.creator is not None and self.creator.is_pi:
+                pi_list.append(self.creator)
+
             for u in self.users:
                 if u.id != self.creator.id:
                     pi_list.append(u)
             return pi_list
 
+
     class Booking(Base):
         """Model for user accounts."""
         __tablename__ = 'bookings'
 
-        id = Column(Integer,
-                    primary_key=True)
+        MAINTENANCE_LIST = ['cycle', 'installation', 'maintenance', 'afis']
+        DEVELOPMENT_LIST = ['method', 'research', 'test', 'mikroed', 'microed', 'devel']
 
-        title = Column(String(256),
-                       nullable=False)
+        id = Column(Integer, primary_key=True)
 
-        start = Column(UtcDateTime,
-                       nullable=False)
+        title = Column(String(256), nullable=False)
 
-        end = Column(UtcDateTime,
-                     nullable=False)
+        start = Column(UtcDateTime, nullable=False)
 
-        # booking, slot or downtime
-        type = Column(String(16),
-                      nullable=False)
+        end = Column(UtcDateTime, nullable=False)
+
+        # booking, slot, downtime or maintenance
+        type = Column(String(16), nullable=False)
 
         # slot authorization, who can book within this slot
         slot_auth = Column(JSON, default={'applications': [], 'users': []})
 
-        description = Column(Text,
-                             nullable=True)
+        description = Column(Text, nullable=True)
 
         repeat_id = Column(String(256), nullable=True)
 
@@ -549,6 +599,11 @@ def create_data_models(dm):
                           nullable=False)
         owner = relationship("User", foreign_keys=[owner_id])
 
+        # And this is the user that "owns" the Booking
+        operator_id = Column(Integer, ForeignKey('users.id'),
+                          nullable=True)
+        operator = relationship("User", foreign_keys=[operator_id])
+
         # Related to the Owner, we also keep the Application to which
         # this booking is associated
         application_id = Column(Integer, ForeignKey('applications.id'),
@@ -660,7 +715,10 @@ def create_data_models(dm):
         end = Column(UtcDateTime)
 
         # Possible statuses of a Session:
-        #   - created (not started)
+        #   - pending (stored in db, but data folders not created)
+        #   - created (data folders created, but no processing reported)
+        # TODO: review if the following states make sense, we might want to
+        # TODO: decouple the session from the associated pre-processing
         #   - running
         #   - failed
         #   - finished
@@ -821,6 +879,166 @@ def create_data_models(dm):
             return dm.json_from_object(self)
 
 
+    class Project(Base):
+        """ Project entity to group shipments, grids preparation,
+         data collections and data processing.
+         """
+        __tablename__ = 'projects'
+
+        id = Column(Integer, primary_key=True)
+
+        date = Column(UtcDateTime, nullable=False)
+
+        STATUS = ['active', 'inactive']
+        # Possible statuses of a Project:
+        #   - active: default state when created
+        #   - inactive: it has been closed and it becomes inactive
+        status = Column(String(32), default='active')
+
+        title = Column(String(256), nullable=False)
+
+        description = Column(Text, nullable=True)
+
+        # User to which this Project is associated with
+        # The PI of this project is inferred
+        user_id = Column(Integer, ForeignKey('users.id'), nullable=False)
+        user = relationship("User", #back_populates= "projects",
+                            foreign_keys=[user_id])
+
+        # Last date (date and user)
+        creation_date = Column(UtcDateTime, nullable=False)
+
+        # Usually it will be facility staff that will update this
+        creation_user_id = Column(Integer, ForeignKey('users.id'),
+                                     nullable=False)
+        creation_user = relationship("User",
+                                        foreign_keys=[creation_user_id])
+
+        # Last date (date and user)
+        last_update_date = Column(UtcDateTime, nullable=False)
+
+        # Usually it will be facility staff that will update this
+        last_update_user_id = Column(Integer, ForeignKey('users.id'),
+                                     nullable=False)
+        last_update_user = relationship("User",
+                                        foreign_keys=[last_update_user_id])
+
+        # General JSON dict to store extra attributes
+        extra = Column(JSON, default={})
+
+        entries = relationship('Entry', back_populates='project')
+
+        def __getExtra(self, key, default):
+            return self.extra.get(key, default)
+
+        def __setExtra(self, key, value):
+            extra = dict(self.extra or {})
+            extra[key] = value
+            self.extra = extra
+
+        @property
+        def user_can_edit(self):
+            """ True if the user of the project can edit it (add/modify/delete notes)
+            """
+            return  self.__getExtra('user_can_edit', False)
+
+        @user_can_edit.setter
+        def user_can_edit(self, value):
+            self.__setExtra('user_can_edit', value)
+
+        def json(self):
+            return dm.json_from_object(self)
+
+
+    class Entry(Base):
+        """ Entry related to a given project.
+         """
+        __tablename__ = 'entries'
+
+        id = Column(Integer, primary_key=True)
+
+        date = Column(UtcDateTime, nullable=False)
+
+        type = Column(String(16), nullable=False)
+
+        title = Column(String(256), nullable=False)
+
+        description = Column(Text, nullable=True)
+
+        # User to which this Project is associated with
+        # The PI of this project
+        project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)
+        project = relationship("Project", back_populates="entries")
+
+        # Last date (date and user)
+        creation_date = Column(UtcDateTime, nullable=False)
+
+        # Usually it will be facility staff that will update this
+        creation_user_id = Column(Integer, ForeignKey('users.id'),
+                                     nullable=False)
+        creation_user = relationship("User",
+                                        foreign_keys=[creation_user_id])
+
+        # Last date (date and user)
+        last_update_date = Column(UtcDateTime, nullable=False)
+
+        # Usually it will be facility staff that will update this
+        last_update_user_id = Column(Integer, ForeignKey('users.id'),
+                                     nullable=False)
+        last_update_user = relationship("User",
+                                        foreign_keys=[last_update_user_id])
+
+        # General JSON dict to store extra attributes
+        extra = Column(JSON, default={})
+
+        def __getExtra(self, key, default):
+            return self.extra.get(key, default)
+
+        def __setExtra(self, key, value):
+            extra = dict(self.extra or {})
+            extra[key] = value
+            self.extra = extra
+
+        def json(self):
+            return dm.json_from_object(self)
+
+
+    class Puck(Base):
+        """ Puck entity for Grids Storage table.
+        """
+        __tablename__ = 'pucks'
+
+        id = Column(Integer,
+                    primary_key=True)
+
+        code = Column(String(64), unique=True, index=True)
+
+        label = Column(String(64),
+                      unique=True,
+                      nullable=False)
+
+        color = Column(String(16))
+
+        # Locations properties
+        dewar = Column(Integer)
+        cane = Column(Integer)
+        position = Column(Integer)
+
+        # General JSON dict to store extra attributes
+        extra = Column(JSON, default={})
+
+        def json(self):
+            return dm.json_from_object(self)
+
+        def __getExtra(self, key, default):
+            return self.extra.get(key, default)
+
+        def __setExtra(self, key, value):
+            extra = dict(self.extra or {})
+            extra[key] = value
+            self.extra = extra
+
+
     dm.Form = Form
     dm.User = User
     dm.Resource = Resource
@@ -830,3 +1048,6 @@ def create_data_models(dm):
     dm.Session = Session
     dm.Transaction = Transaction
     dm.InvoicePeriod = InvoicePeriod
+    dm.Project = Project
+    dm.Entry = Entry
+    dm.Puck = Puck
