@@ -26,6 +26,7 @@
 
 import json
 import datetime as dt
+import numpy as np
 
 from . import image
 
@@ -38,6 +39,9 @@ def pretty_date(input_dt):
     if input_dt is None:
         return 'None'
 
+    if isinstance(input_dt, str):
+        input_dt = datetime_from_isoformat(input_dt)
+
     return input_dt.strftime("%Y/%m/%d")
 
 
@@ -45,12 +49,15 @@ def pretty_datetime(input_dt):
     if input_dt is None:
         return 'None'
 
+    if isinstance(input_dt, str):
+        input_dt = datetime_from_isoformat(input_dt)
+
     return input_dt.strftime("%Y/%m/%d %I:%M %p")
 
 
 def pretty_quarter(quarter):
     qs, qe = quarter
-    return '%s - %s' % (qs.strftime('%b'),
+    return '%s - %s' % (qs.strftime('%b %Y'),
                         qe.strftime('%b %Y'))
 
 
@@ -88,9 +95,20 @@ def get_quarter(date=None):
     return _dt(start), _dt(end)
 
 
+class NpJsonEncoder(json.JSONEncoder):
+    def default(self, obj):
+        if isinstance(obj, np.integer):
+            return int(obj)
+        if isinstance(obj, np.floating):
+            return float(obj)
+        if isinstance(obj, np.ndarray):
+            return obj.tolist()
+        return super(NpJsonEncoder, self).default(obj)
+
+
 def send_json_data(data):
     import flask
-    resp = flask.make_response(json.dumps(data))
+    resp = flask.make_response(json.dumps(data, cls=NpJsonEncoder))
     resp.status_code = 200
     resp.headers['Access-Control-Allow-Origin'] = '*'
     return resp
